#ifndef H_TILEMAP
#define H_TILEMAP
 
#include <array>
#include <SFML/Graphics.hpp>
 
class TexBoxCoords : public sf::Rect<int> {
public:
	TexBoxCoords(void) :
		sf::Rect<int>()
	{}
	TexBoxCoords(const sf::Rect<int>& rectangle) :
		sf::Rect<int>(rectangle)
	{}
	TexBoxCoords(const sf::Vector2<int>& position, const sf::Vector2<int>& size) :
		Rect<int>(position, size)
	{}
 
	sf::Vector2f TopLeft(void) const {
		return sf::Vector2f(left, top);
	}
	sf::Vector2f TopRight(void) const {
		return sf::Vector2f(left + width, top);
	}
	sf::Vector2f BottomLeft(void) const {
		return sf::Vector2f(left, top + height);
	}
	sf::Vector2f BottomRight(void) const {
		return sf::Vector2f(left + width, top + height);
	}
};
 
template <int resX, int resY>
class TileMap : public sf::Drawable, public sf::Transformable {
public:
	TileMap(sf::Vector2f size, sf::Texture* _texture, std::vector<TexBoxCoords>& texCoords, std::array<std::array<int, resX>, resY>& mapData) :
		tile_size(sf::Vector2i(1 / resX * size.x, 1 / resY * size.y)),
		vertex_count(resX * resY * 4),
		texture(_texture),
		vertices(sf::VertexArray(sf::Quads, 0u))
	{
		for(int y = 0; y < resY; y++){
			for(int x = 0; x < resX; x++){
				sf::Vertex vertex0(sf::Vector2f((float) x / (float) resX * size.x, (float) y / (float) resY * size.y), sf::Color::White, texCoords[mapData[y][x]].TopLeft());
				sf::Vertex vertex1(sf::Vector2f((float) (x + 1) / (float) resX * size.x, (float) y / (float) resY * size.y), sf::Color::White, texCoords[mapData[y][x]].TopRight());
				sf::Vertex vertex2(sf::Vector2f((float) (x + 1) / (float) resX * size.x, (float) (y + 1) / (float) resY * size.y), sf::Color::White, texCoords[mapData[y][x]].BottomRight());
				sf::Vertex vertex3(sf::Vector2f((float) x / (float) resX * size.x, (float) (y + 1) / (float) resY * size.y), sf::Color::White, texCoords[mapData[y][x]].BottomLeft());
 
				vertices.append(vertex0);
				vertices.append(vertex1);
				vertices.append(vertex2);
				vertices.append(vertex3);
			}
		}
	}
 
	sf::Vector2i getTileSize(void) const {
		return tile_size * getScale();
	}
 
private:
	virtual void draw(sf::RenderTarget& target, sf::RenderStates states) const {
		states.texture = texture;
		states.transform *= getTransform();
 
		target.draw(vertices, states);
	}
 
	sf::Vector2i tile_size;
	unsigned int vertex_count;
	sf::VertexArray vertices;
	sf::Texture* texture;
};
 
#endif