#include <array>
#include <fstream>
#include <iostream>
#include <memory>
#include <SFML/Graphics.hpp>
#include <sstream>
#include "TileMap.h"
#include <vector>

struct Obj {
	int X;
	int Y;
	int layer;
	bool canGrab;
	bool isHeld;
	int destination;

	int special;
};

bool isInteger(char);
void parseTileString(std::array<std::array<int, 21>, 16>&, std::string);
void collideAndRespond(float&, float&, float&, bool&, std::array<std::array<int, 21>, 16>);
std::vector<Obj> loadInObjects(std::ifstream&);

int main(void){
	std::vector<Obj> objTable;

	bool isTransitioning = false;
	int transitionTimer;

	bool isPaused = false;

	int worldView = 0;

	int currentLevel = 1;

	int gX;
	int gY;
	int goalLayer;

	int frameCount = 0;
	std::string tmpTileLine;
	std::ifstream tilesIn;
	tilesIn.open("TS3DL1.txt");

	sf::Texture endTexture;{
		if(!endTexture.create(512, 384))
			return 1;
		endTexture.loadFromFile("endScreen1.png");
	}

	sf::Sprite endScreen(endTexture);

	sf::Texture travelTex;{
		if(!travelTex.create(96, 72))
			return 1;
		travelTex.loadFromFile("travelPointSprite2.png");
	}

	sf::Sprite travelSprite(travelTex, sf::IntRect(0,0, 24,24));

	//Player Status. OOPify later
	float pX = 0;
	float pY = 0;
	float pVY = 0;
	float pX2 = pX / 24;
	float pY2 = pY / 24;
	bool canJump = false;
	bool canMove = true;
	bool canTravel = true;
	bool isHolding = false;
	int dazeTimer = 0;
	int travelTimer = 0;
	int holdTimer = 0;
	int dropTimer = 0;
	int resetTimer = 0;
	Obj* holdObj;

	sf::View view(sf::Vector2f(-4.f + 256.f, 192.f), sf::Vector2f(512.f, 384.f));
	sf::RenderWindow window(sf::VideoMode(512u, 384u), "Coplanarity");{
		window.setFramerateLimit(60);
		window.setView(view);
	}

	sf::RectangleShape transitionRect(sf::Vector2f(504, 384));

	sf::Texture tileSheet;{
		if(!tileSheet.create(48,24))
			return 1;
		tileSheet.loadFromFile("tooSpooky3DTileset3.png");
	}
	std::vector<TexBoxCoords> tileTexCoords;{
		tileTexCoords.push_back(TexBoxCoords(sf::Vector2i(48, 24), sf::Vector2i(0, 0)));
		for(int yi = 0; yi < 4; ++yi){
			for(int xi = 0; xi < 4; ++xi){
				tileTexCoords.push_back(TexBoxCoords(sf::Vector2i(xi*24, yi*24), sf::Vector2i(24, 24)));
			}
		}
	}
	std::array<std::array<int, 21>, 16> backgroundDisplay;
	std::array<std::array<int, 21>, 16> frontDisplay;
	std::array<std::array<int, 21>, 16> frontCollide;
	std::array<std::array<int, 21>, 16> midDisplay;
	std::array<std::array<int, 21>, 16> midCollide;
	std::array<std::array<int, 21>, 16> deepDisplay;
	std::array<std::array<int, 21>, 16> deepCollide;{
		if(tilesIn.is_open()){
			//Background
			std::getline(tilesIn, tmpTileLine);
			parseTileString(backgroundDisplay, tmpTileLine);
			//Front
			//Display tiles
			std::getline(tilesIn, tmpTileLine);
			parseTileString(frontDisplay, tmpTileLine);
			//Collision tiles
			std::getline(tilesIn, tmpTileLine);
			parseTileString(frontCollide, tmpTileLine);
			//Mid
			//Display tiles
			std::getline(tilesIn, tmpTileLine);
			parseTileString(midDisplay, tmpTileLine);
			//Collision tiles
			std::getline(tilesIn, tmpTileLine);
			parseTileString(midCollide, tmpTileLine);
			//Deep
			//Display tiles
			std::getline(tilesIn, tmpTileLine);
			parseTileString(deepDisplay, tmpTileLine);
			//Collision tiles
			std::getline(tilesIn, tmpTileLine);
			parseTileString(deepCollide, tmpTileLine);

			//Player Pos
			std::getline(tilesIn, tmpTileLine);
			pX = std::stoi(tmpTileLine) * 24;
			std::getline(tilesIn, tmpTileLine);
			pY = std::stoi(tmpTileLine) * 24;

			//Goal Pos
			std::getline(tilesIn, tmpTileLine);
			gX = std::stoi(tmpTileLine) * 24;
			std::getline(tilesIn, tmpTileLine);
			gY = std::stoi(tmpTileLine) * 24;
			std::getline(tilesIn, tmpTileLine);
			goalLayer = std::stoi(tmpTileLine);
			
			//Load in objects
			objTable = loadInObjects(tilesIn);
		}
		tilesIn.close();
	}
	/* = {
		0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
		0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
		0,0,1,0,0,1,0,0,1,1,0,0,0,0,1,1,0,0,0,0,0,
		0,0,1,0,0,1,0,0,1,1,0,0,0,0,1,1,0,0,0,0,0,
		0,0,0,0,0,0,0,0,1,1,0,0,0,0,1,1,0,0,0,0,0,
		0,0,1,1,1,1,0,0,1,1,0,0,0,0,1,1,0,0,0,0,0,
		0,1,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
		0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
		0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
		0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,1,1,0,0,0,
		0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,1,1,0,0,0,
		0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,
		0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,
		0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
		0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
		0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	};*/
	std::unique_ptr<TileMap<21, 16>> backmap(new TileMap<21, 16>(sf::Vector2f(504.f, 384.f), &tileSheet, tileTexCoords, backgroundDisplay));
	std::unique_ptr<TileMap<21, 16>> frontmap(new TileMap<21, 16>(sf::Vector2f(504.f, 384.f), &tileSheet, tileTexCoords, frontDisplay));
	std::unique_ptr<TileMap<21, 16>> midmap(new TileMap<21, 16>(sf::Vector2f(504.f, 384.f), &tileSheet, tileTexCoords, midDisplay));
	std::unique_ptr<TileMap<21, 16>> deepmap(new TileMap<21, 16>(sf::Vector2f(504.f, 384.f), &tileSheet, tileTexCoords, deepDisplay));
	
	sf::Texture plyrTex;{
		if(!plyrTex.create(288, 48))
			return 1;
		plyrTex.loadFromFile("playerSpriteTS3D1.png");
	}
	sf::Sprite plyrSprite(plyrTex, sf::IntRect(0,0, 48,48));{
		plyrSprite.setScale(0.5f, 0.5f);
		plyrSprite.setOrigin(0.f, 0.f);
	}

	while(window.isOpen()){
		sf::Event event;
		while(window.pollEvent(event)){
			switch(event.type){
			case sf::Event::Closed:
				window.close();
				break;
			}
		}
		if(!isTransitioning){
			plyrSprite.setTextureRect(sf::IntRect(0,0, 48,48));
			if(sf::Keyboard::isKeyPressed(sf::Keyboard::Left) && canMove){
				plyrSprite.setScale(-0.5, 0.5);
				plyrSprite.setOrigin(48, 0);
				pX -= 5.f/3.f;
				plyrSprite.setTextureRect(sf::IntRect(96,0, 48,48));
				if(frameCount % 12 <= 3){
					plyrSprite.setTextureRect(sf::IntRect(48,0, 48,48));
				}
			}else
			if(sf::Keyboard::isKeyPressed(sf::Keyboard::Right) && canMove){
				plyrSprite.setScale(0.5, 0.5);
				plyrSprite.setOrigin(0, 0);
				pX += 5.f/3.f;
				plyrSprite.setTextureRect(sf::IntRect(96,0, 48,48));
				if(frameCount % 12 <= 3){
					plyrSprite.setTextureRect(sf::IntRect(48,0, 48,48));
				}
			}
			if(sf::Keyboard::isKeyPressed(sf::Keyboard::Up) && canJump && canMove){
				pVY = -5.f;
			}
			if(sf::Keyboard::isKeyPressed(sf::Keyboard::Z) && canTravel && !isHolding){
				Obj* travelObj;
				bool closeEnough = false;
				for(std::vector<Obj>::iterator it = objTable.begin(); it != objTable.end(); ++it){
					if(abs((*it).X - pX) < 24 && abs((*it).Y - pY) < 24 && (*it).layer == worldView && (*it).special == 2){
						travelObj = &(*it);
						closeEnough = true;
						//break;
					}
				}
				if(closeEnough){
					int tmp;
					travelTimer = frameCount;
					canTravel = false;
					transitionTimer = frameCount;
					isTransitioning = true;
					tmp = worldView;
					worldView = travelObj->destination;
					travelObj->layer = worldView;
					travelObj->destination = tmp;
				}
			}
			if(sf::Keyboard::isKeyPressed(sf::Keyboard::X) && !isHolding && frameCount - dropTimer > 16){
				bool closeEnough = false;
					for(std::vector<Obj>::iterator it = objTable.begin(); it != objTable.end(); ++it){
					if(abs((*it).X - pX) < 24 && abs((*it).Y - pY) < 24 && (*it).layer == worldView && (*it).canGrab){
						holdTimer = frameCount;
						holdObj = &(*it);
						isHolding = true;
						break;
					}
				}
			}
			if(sf::Keyboard::isKeyPressed(sf::Keyboard::X) && isHolding && frameCount - holdTimer > 16){
				dropTimer = frameCount;
				isHolding = false;
				holdObj->X = pX;
				holdObj->Y = pY;
				holdObj = NULL;
			}
			if(sf::Keyboard::isKeyPressed(sf::Keyboard::R) && frameCount - resetTimer > 16){
				worldView = 0;
				resetTimer = frameCount;
				isHolding = false;
				holdObj = NULL;
				objTable.clear();
				std::stringstream strs;
				strs << "TS3DL" << currentLevel << ".txt";
				std::ifstream tilesIn2(strs.str());
				if(tilesIn2.is_open()){
					//Background
					std::getline(tilesIn2, tmpTileLine);
					parseTileString(backgroundDisplay, tmpTileLine);
					//Front
					//Display tiles
					std::getline(tilesIn2, tmpTileLine);
					parseTileString(frontDisplay, tmpTileLine);
					//Collision tiles
					std::getline(tilesIn2, tmpTileLine);
					parseTileString(frontCollide, tmpTileLine);
					//Mid
					//Display tiles
					std::getline(tilesIn2, tmpTileLine);
					parseTileString(midDisplay, tmpTileLine);
					//Collision tiles
					std::getline(tilesIn2, tmpTileLine);
					parseTileString(midCollide, tmpTileLine);
					//Deep
					//Display tiles
					std::getline(tilesIn2, tmpTileLine);
					parseTileString(deepDisplay, tmpTileLine);
					//Collision tiles
					std::getline(tilesIn2, tmpTileLine);
					parseTileString(deepCollide, tmpTileLine);

					//Player Pos
					std::getline(tilesIn2, tmpTileLine);
					pX = std::stoi(tmpTileLine) * 24;
					std::getline(tilesIn2, tmpTileLine);
					pY = std::stoi(tmpTileLine) * 24;

					//Goal Pos
					std::getline(tilesIn2, tmpTileLine);
					gX = std::stoi(tmpTileLine) * 24;
					std::getline(tilesIn2, tmpTileLine);
					gY = std::stoi(tmpTileLine) * 24;
					std::getline(tilesIn2, tmpTileLine);
					goalLayer = std::stoi(tmpTileLine);

					//Load in objects
					objTable = loadInObjects(tilesIn2);
				}
				tilesIn2.close();

				backmap = std::unique_ptr<TileMap<21,16>>(new TileMap<21, 16>(sf::Vector2f(504.f, 384.f), &tileSheet, tileTexCoords, backgroundDisplay));
				frontmap = std::unique_ptr<TileMap<21,16>>(new TileMap<21, 16>(sf::Vector2f(504.f, 384.f), &tileSheet, tileTexCoords, frontDisplay));
				midmap = std::unique_ptr<TileMap<21,16>>(new TileMap<21, 16>(sf::Vector2f(504.f, 384.f), &tileSheet, tileTexCoords, midDisplay));
				deepmap = std::unique_ptr<TileMap<21,16>>(new TileMap<21, 16>(sf::Vector2f(504.f, 384.f), &tileSheet, tileTexCoords, deepDisplay));
			}

			if(isHolding){
				holdObj->X = pX;
				holdObj->Y = pY-24;
			}

			if(!canTravel && frameCount - travelTimer > 16){
				canTravel = true;
			}

			travelSprite.setTextureRect(sf::IntRect((frameCount/3%4)*24,0, 24,24));

			if(!canJump && pVY < 8){
				plyrSprite.setTextureRect(sf::IntRect(96,0, 48,48));
			}else
			if(!canJump){
				plyrSprite.setTextureRect(sf::IntRect(144,0, 48,48));
			}

			pY += pVY;
			pVY += pVY < 8 ? 3.f/8.f : 0;

			if(pX < 0)
				pX = 0;
			if(pX > 504 - 24)
				pX = 504 - 24;
			if(pY > 384 - 24)
				pY = 384 - 24;

			int savedV = pVY;
			switch(worldView){
			case 0:
				collideAndRespond(pX, pY, pVY, canJump, frontCollide);
			break;
			case 1:
				collideAndRespond(pX, pY, pVY, canJump, midCollide);
			break;
			case 2:
				collideAndRespond(pX, pY, pVY, canJump, deepCollide);
			break;
			}
			if(savedV >= 8 && canJump){
				canMove = false;
				dazeTimer = frameCount;
			}

			if(!canMove){
				if(frameCount % 16 <= 8)
					plyrSprite.setTextureRect(sf::IntRect(192,0, 48,48));
				else
					plyrSprite.setTextureRect(sf::IntRect(240,0, 48,48));
				if(frameCount - dazeTimer >= 60){
					canMove = true;
				}
			}

			if(abs(gX - pX) < 24 && abs(gY - pY) < 24 && worldView == goalLayer){
				isHolding = false;
				holdObj = NULL;
				objTable.clear();
				currentLevel++;
				std::stringstream strs;
				strs << "TS3DL" << currentLevel << ".txt";
				std::ifstream tilesIn2(strs.str());
				if(tilesIn2.is_open()){
					//Background
					std::getline(tilesIn2, tmpTileLine);
					parseTileString(backgroundDisplay, tmpTileLine);
					//Front
					//Display tiles
					std::getline(tilesIn2, tmpTileLine);
					parseTileString(frontDisplay, tmpTileLine);
					//Collision tiles
					std::getline(tilesIn2, tmpTileLine);
					parseTileString(frontCollide, tmpTileLine);
					//Mid
					//Display tiles
					std::getline(tilesIn2, tmpTileLine);
					parseTileString(midDisplay, tmpTileLine);
					//Collision tiles
					std::getline(tilesIn2, tmpTileLine);
					parseTileString(midCollide, tmpTileLine);
					//Deep
					//Display tiles
					std::getline(tilesIn2, tmpTileLine);
					parseTileString(deepDisplay, tmpTileLine);
					//Collision tiles
					std::getline(tilesIn2, tmpTileLine);
					parseTileString(deepCollide, tmpTileLine);

					//Player Pos
					std::getline(tilesIn2, tmpTileLine);
					pX = std::stoi(tmpTileLine) * 24;
					std::getline(tilesIn2, tmpTileLine);
					pY = std::stoi(tmpTileLine) * 24;

					//Goal Pos
					std::getline(tilesIn2, tmpTileLine);
					gX = std::stoi(tmpTileLine) * 24;
					std::getline(tilesIn2, tmpTileLine);
					gY = std::stoi(tmpTileLine) * 24;
					std::getline(tilesIn2, tmpTileLine);
					goalLayer = std::stoi(tmpTileLine);

					//Load in objects
					objTable.clear();
					objTable = loadInObjects(tilesIn2);
				}
				tilesIn2.close();

				backmap = std::unique_ptr<TileMap<21,16>>(new TileMap<21, 16>(sf::Vector2f(504.f, 384.f), &tileSheet, tileTexCoords, backgroundDisplay));
				frontmap = std::unique_ptr<TileMap<21,16>>(new TileMap<21, 16>(sf::Vector2f(504.f, 384.f), &tileSheet, tileTexCoords, frontDisplay));
				midmap = std::unique_ptr<TileMap<21,16>>(new TileMap<21, 16>(sf::Vector2f(504.f, 384.f), &tileSheet, tileTexCoords, midDisplay));
				deepmap = std::unique_ptr<TileMap<21,16>>(new TileMap<21, 16>(sf::Vector2f(504.f, 384.f), &tileSheet, tileTexCoords, deepDisplay));

			}
		}

		plyrSprite.setPosition(pX, pY);
		window.clear();
		window.draw(*backmap);
		switch(worldView){
		case 0:
			window.draw(*frontmap);
		break;
		case 1:
			window.draw(*midmap);
		break;
		case 2:
			window.draw(*deepmap);
		break;
		}
		window.draw(plyrSprite);
		for(std::vector<Obj>::iterator it = objTable.begin(); it != objTable.end(); ++it){
			if(it->layer == worldView){
				if(it->special == 2){
					travelSprite.setPosition(it->X, it->Y);
					travelSprite.setTextureRect(sf::IntRect(travelSprite.getTextureRect().left, it->destination * 24, 24, 24));
					window.draw(travelSprite);
				}
			}
		}
		if(isTransitioning){
			if(frameCount - transitionTimer < 64){
				transitionRect.setFillColor(sf::Color(255,255,255,255 - (frameCount - transitionTimer)*255/64));
				window.draw(transitionRect);
			}else
				isTransitioning = false;
		}
		if(currentLevel > 24){
			window.draw(endScreen);
		}
		window.display();
		frameCount++;
	}
	return 0;
}

bool isInteger(char num){
	//Is it within the range of numbers in ASCII?
	return (int)num >= 48 && (int)num <= 57;
}

void parseTileString(std::array<std::array<int, 21>, 16>& mapData, std::string tileString){
	std::string numberHolder;
	int numberCount = 0;
	for(std::string::iterator it = tileString.begin(); it != tileString.end(); ++it){
		if(isInteger(*it)){
			numberHolder.push_back(*it);
		}else{
			if(numberHolder.length() > 0){
				mapData[numberCount / 21][numberCount % 21] = std::stoi(numberHolder);
				numberCount++;
				numberHolder.clear();
			}
		}
	}
	mapData[15][20] = std::stoi(numberHolder);
	tileString.clear();
}

void collideAndRespond(float& oX, float& oY, float& pVY, bool& canJump, std::array<std::array<int, 21>, 16> collideMap){
	float oX2 = oX / 24;
	float oY2 = oY / 24;
	canJump = false;
	for(float colX = floor(oX2 - 1); colX < floor(oX2 + 2) + 0.5f; ++colX){
		for(float colY = floor(oY2 - 1); colY < floor(oY2 + 2) + 0.5f; ++colY){
			if(colX >= 0 && colX < 21 && colY >= 0 && colY < 16 && collideMap[colY][colX] > 0){
				if(abs(oX2 - colX) < 1 && abs(oY2 - colY) < 1){
					if(0.9 * abs(cos(atan2(oY2 - colY,oX2 - colX))) > abs(sin(atan2(oY2 - colY,oX2 - colX)))){
						if(oX2 - colX > 0){
							oX = (colX + 1) * 24;
						}else
						if(oX2 - colX < 0){
							oX = (colX - 1) * 24;
						}
					}else
					if(0.9 * abs(sin(atan2(oY2 - colY,oX2 - colX))) > abs(cos(atan2(oY2 - colY,oX2 - colX)))){
						if(oY2 - colY > 0){
							oY = (colY + 1) * 24;
							pVY = 0;
						}else
						if(oY2 - colY < 0){
							oY = (colY - 0.99) * 24;
							pVY = 0;
							canJump = true;
						}
					}
				}
			}
		}
	}
}

std::vector<Obj> loadInObjects(std::ifstream& file){
	std::string line;
	std::string sub;
	int objectCount;
	std::getline(file, line);
	objectCount = std::stoi(line);
	std::vector<Obj> objTable(objectCount);
	for(int j = 0; j < objTable.size(); ++j){
		int spacePos;
		int deadPos = 0;
		int wordCount = 0;
		std::getline(file, line);
		while(wordCount <= 5){
			for(int i = deadPos; i < line.length(); ++i){
				if(line[i] == ' '){
					spacePos = i;
					switch(wordCount){
					case 0:
						objTable[j] = Obj();
						objTable[j].special = 1 + std::stoi(line.substr(deadPos, spacePos - deadPos));
						objTable[j].isHeld = false;
					break;
					case 1:
						sub = line.substr(deadPos+1, spacePos - deadPos);
						objTable[j].X = 24 * std::stoi(sub);
					break;
					case 2:
						sub = line.substr(deadPos+1, spacePos - deadPos);
						objTable[j].Y = 24 * std::stoi(sub);
					break;
					case 3:
						sub = line.substr(deadPos+1, spacePos - deadPos);
						objTable[j].layer = std::stoi(sub);
					break;
					case 4:
						sub = line.substr(deadPos+1, spacePos - deadPos);
						objTable[j].canGrab = (bool)std::stoi(sub);
					break;
					case 5:
						sub = line.substr(deadPos+1, spacePos - deadPos);
						objTable[j].destination = std::stoi(sub);
					break;
					}
					wordCount++;
					deadPos = spacePos;
				}
			}
		}
	}
	return objTable;
}